/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mintbind.h>
#include <mint/cookie.h>

#include "general.h"

#include "av.h"
#include "manager.h"
#include "utils.h"
#include "protocols.h"

/**************************************************************************************************/
/* AV Protocol																					  */
/**************************************************************************************************/
/*
 * Declare ourself to the AV server. Copy formatted name into buffer.
 */
int init_av_protokoll(int os, char *name)
{
	int found        = FALSE;
	int mode         = MX_GLOBAL | MX_PREFTTRAM;
	char *client	 = NULL;
	char avserver[9] = {0};

	if (os == MINT) {
		client = (char*)Mxalloc(9, mode);
	}
	else if (os == MAGIC) {
		client = (char*)Mxalloc(9, (int)mode & 0xFF);
	}

	if (client != NULL)
	{
		short msg[8] = {AV_PROTOKOLL, _AESapid, 0, 0x02, 0, 0, 0, 0};
		char *server = NULL;

		strcpy(client, "GEMCLIP "); // Must be 8 characters long.
		shel_envrn(&server, "AVSERVER=");

		if (server != NULL)
		{
			sprintf(avserver, "%-8s", server); // Reformat name to be GEMDOS-compliant.
			*(char **)&msg[6] = client;
			av_protokoll_send_message(avserver, msg);
			found = TRUE;
			strcpy(name, avserver);
		}
		else {
			Mfree(client);
		}
	}

	return found;
}

/*
 * Tell the AV server we are quitting.
	 */
void av_protokoll_exit(const char *server)
{
	short msg[8] = {AV_EXIT, 0, 0, _AESapid, 0, 0, 0, 0};

	av_protokoll_send_message(server, msg);
}

/*
 * Report opening/closing of windows to the AV server.
 */
void av_protokoll_window(short handle, const char *server, int message)
{
	short msg[8] = {message, _AESapid, 0, handle, 0, 0, 0, 0};

	av_protokoll_send_message(server, msg);
}

/*
 * Send Ctrl+w to the AV server for global window cycling.
 */
void av_protokoll_send_key(const char *server, short kreturn)
{
	short msg[8] = {AV_SENDKEY, _AESapid, 0, K_CTRL, (kreturn << 8), 0, 0, 0};

	av_protokoll_send_message(server, msg);
}

/*
 * Ask the AV server to open a file with a possible command line (may be NULL).
 */
void av_protokoll_open_file(const char *path, int size, const char *command, const char *server, int os)
{
	int mode      = MX_GLOBAL | MX_PREFTTRAM;
	short msg[8]  = {AV_STARTPROG, _AESapid, 0, 0, 0, 0, 0, 0};
	char *global1 = NULL;
	char *global2 = NULL;

	if (os == MINT) {
		global1 = (char*)Mxalloc(size, mode);
	}
	else if (os == MAGIC) {
		global1 = (char*)Mxalloc(size, (int)mode & 0xFF);
	}
	
	if (command != NULL)
	{
		if (os == MINT) {
			global2 = (char*)Mxalloc(LBUFFER, mode);
		}
		else if (os == MAGIC) {
			global2 = (char*)Mxalloc(LBUFFER, (int)mode & 0xFF);
		}
	}
	
	if (global1 != NULL)
	{
		strcpy(global1, path);
		
		if ((command != NULL) && (global2 != NULL))
		{
			strcpy(global2, command);
			*(char**)(&msg[5]) = global2;
		}

		*(char**)(&msg[3]) = global1;
		av_protokoll_send_message(server, msg);
	}
}

#if 0
/*
 * Ask the AV server to open a file. NOTE: unused.
 */
void av_protokoll_delete_file(const char *path, int size, const char *server)
{
	short msg[8] = {AV_DELFILE, _AESapid, 0, 0, 0, 0, 0, 0};
	char *global = NULL;

	if ((global = (char*)Mxalloc(size, MX_PREFTTRAM | MX_GLOBAL)) != NULL)
	{
		strcpy(global, path);
		*(char**)(&msg[3]) = global;
		av_protokoll_send_message(server, msg);
	}
}
#endif

/*
 * Ask the AV server to update a directory window.
 */
void av_protokoll_path_update(const char *server, const char *path, int os)
{
	int mode     = MX_GLOBAL | MX_PREFTTRAM;
	short msg[8] = {AV_PATH_UPDATE, _AESapid, 0, 0, 0, 0, 0, 0};
	char *global = NULL;

	/* Malloc in global mode. */
	if (os == MINT) {
		global = (char*)Mxalloc(strlen(path) + 1, mode);
	}
	else if (os == MAGIC) {
		global = (char*)Mxalloc(strlen(path) + 1, (int)mode & 0xFF);
	}

	if (global != NULL)
	{
		strcpy(global, path);
		*(char**)(&msg[3]) = global;
		av_protokoll_send_message(server, msg);
	}
}


/*
 * Ask the AV server to update the parent directory of the directory found in path.
 * Example: if 'path' == "C:\\CLIPBRD", update 'C:\\'
 */
void update_desktop_window(char *path, const char *avserver, int os)
{
	char *p = NULL;

	if ((p = strrchr(path, '\\')) != NULL)
	{
		while (*--p != '\\')
			;
	
		*(p + 1) = '\0';
	}

	av_protokoll_path_update(avserver, path, os);
}

#if 0
/*
 * Ask the AV server to update the clipboard directory. NOTE: unused.
 */
void av_protokoll_clipbrd_update(const char *server)
{
	short msg[8] = {AV_PATH_UPDATE, _AESapid, 0, 0, 0, 0, 0, 0};
	char *path	 = NULL;
	char scrapdir[MBUFFER] = {0};

	if (scrp_read(scrapdir) != 0)
	{
		/* Malloc in global mode. */
		path = (char*)Mxalloc(sizeof scrapdir, MX_PREFTTRAM | MX_GLOBAL);

		if (path != NULL)
		{
			strcpy(path, scrapdir);

			/* It seems that we must convert the path to UPPERCASE? */
			for (char *p = path; *p; p++) {
				*p = toupper(*p);
			}

			*(char**)(&msg[3]) = path;
			av_protokoll_send_message(server, msg);
		}
	}
}
#endif

/*
 * Send the message to the AV server. NOTE: shouldn't we free memory?
 */
void av_protokoll_send_message(const char *avserver, short msg[])
{
	short id = -1;

	if ((id = appl_find(avserver)) >= 0)
	{
		appl_write(id, 16, msg);
	}
}

/**************************************************************************************************/
/* BubbleGEM protocol																			  */
/**************************************************************************************************/
/*
 * Process the file containing the BubbleGEM text messages ('bubbles.bub').
 * We are not particularly conservative here and allocate NBUBBLES * LBUFFER
 * straight on, rather that strlen'ing each string on return. I think this is quite
 * acceptable.
 * ==> Return FALSE on error.
 */
#define NBUBBLES  46 // The number of BubbleGEM help messages.

int init_bubbles(BUBBLES *bubbles, int aeslang, int os)
{
	char *env  = NULL;
	char path[LBUFFER] = {0};
	int bubok  = FALSE;
	int idx    = 0;
	int loop   = TRUE;
	int rtrn   = TRUE;
	int mode   = MX_GLOBAL | MX_PREFTTRAM;
	size_t len = 0;
	FILE *fp   = NULL;

	/* The left column keys of the 'gemdict.bub' file. */
	const char *Key[] = { "MGRCURR", "MGRVIEW", "MGRSAVE", "MGRDELE", "MGRMODE", "MGRTEXT",
						  "MGRINFO", "MGRNAME", "MGRDATE", "MGRSIZE", "MGRORDR", "MGRICON",
						  "MGRSCRP", "MGRLIST", "MGRPREV",
						  
						  "ABTOKAY", "ABTSITE", "ABTHELP",
						  
						  "CNFMISC", "CNFKEYS", "CNFLIMT", "CNFICON", "CNFAUTO", "CNFSETT",
						  "CNFCYCL", "CNFBUBL", "CNFMPTY", "CNFWRNG", "CNFDBCK", "CNFRIGT",
						  "CNFRTRN", "CNFSPAC", "CNFINSR", "CNFSIZE", "CNFCLIP", "CNFFREE",
						  "CNFPATH", "CNFOVER", "CNFOKAY", "CNFCNCL", "CNFAPLY", "CNFDFLT",
						  "CNFHELP",
						  
						  "INFPREV", "INFNEXT", "INFHELP" };


	shel_envrn(&env, "ACCPATH=");

	/* Load bubble file according to system locale. We look about everywhere sensible just
	   to be sure (e.g. with TeraDesk). */
	if (aeslang == AESLANG_GERMAN)
	{
		if (env != NULL) {
			snprintf(path, sizeof path, "%s%s%s", env, "\\", "GEMCLIP\\GEMCLIPD.BUB");
		}
		
		if (((fp = fopen("gemclip\\gemclipd.bub", "r"))		     != NULL) ||
			((fp = fopen("GEMCLIP\\GEMCLIPD.BUB", "r")) 		 != NULL) ||
			((fp = fopen("gemclipd.bub", "r"))					 != NULL) ||
			((fp = fopen("GEMCLIPD.BUB", "r"))					 != NULL) ||
			((fp = fopen("C:\\GEMCLIP\\GEMCLIPD.BUB", "r")) 	 != NULL) ||
			((fp = fopen("C:\\ACC\\GEMCLIP\\GEMCLIPD.BUB", "r")) != NULL) ||
			((fp = fopen(path, "r"))                             != NULL))
		{
			bubok = TRUE;
		}
	}
	else
	{
		if (env != NULL) {
			snprintf(path, sizeof path, "%s%s%s", env, "\\", "GEMCLIP\\GEMCLIPE.BUB");
		}

		if (((fp = fopen("gemclip\\gemclipe.bub", "r"))		     != NULL) ||
			((fp = fopen("GEMCLIP\\GEMCLIPE.BUB", "r")) 		 != NULL) ||
			((fp = fopen("gemclipe.bub", "r"))					 != NULL) ||
			((fp = fopen("GEMCLIPE.BUB", "r"))					 != NULL) ||
			((fp = fopen("C;\\GEMCLIP\\GEMCLIPE.BUB", "r")) 	 != NULL) ||
			((fp = fopen("C;\\ACC\\GEMCLIP\\GEMCLIPE.BUB", "r")) != NULL) ||
			((fp = fopen(path, "r"))                             != NULL))
		{
			bubok = TRUE;
		}
	}

	if (bubok == TRUE)
	{
		if (fgetc(fp) != EOF)
		{
			if (os == MINT)
			{
				bubbles->p = (char**)Mxalloc(NBUBBLES * (sizeof *bubbles->p), mode);
				bubbles->buffer = (char*)Mxalloc(NBUBBLES * LBUFFER, mode);
			}
			else if (os == MAGIC)
			{
				bubbles->p = (char**)Mxalloc(NBUBBLES * (sizeof *bubbles->p), (int)mode & 0xFF);
				bubbles->buffer = (char*)Mxalloc(NBUBBLES * LBUFFER, (int)mode & 0xFF);
			}

			if ((bubbles->p != NULL) && (bubbles->buffer != NULL))
			{
				/* Look up a key in the file, copy the value into a globally malloc'd buffer. */
				while ((loop == TRUE) && (idx < NBUBBLES))
				{
					if ((get_value(fp, Key[idx], &bubbles->buffer[len], LBUFFER)) != NULL)
					{
						if (&bubbles->buffer[len] != NULL)
						{
							bubbles->p[idx] = &bubbles->buffer[len];
							len += LBUFFER;
						}
						else {
							loop = FALSE;
						}
					}
					else {
						bubbles->p[idx] = '\0'; // Empty line in case of error.
					}
					idx++;
				}
			}
			else { // FIXME: set to FALSE.
				form_alert(1, FA_ERROR "[error Malloc'ing in|init_bubbles().][OK]");
			}

			if (fclose(fp) == EOF) {
				form_alert(1, FA_ERROR "[Error closing bubble file.][OK]");
			}
		}
		else
		{
			form_alert(1, FA_ERROR "[Error fgetc'ing in|init_bubbles().][OK]");
			rtrn = FALSE;
		}
	}
	else // No way to find that damn file. Exit.
	{
		form_alert(1, FA_ERROR "[Error fopen'ing file|in init_bubbles().][OK]");
		rtrn = FALSE;
	}

	return rtrn;
}

/*
 * Free the memory allocated for the BubbleGEM help messages.
 */
void free_bubbles(BUBBLES *bubbles, USERCONF *conf)
{
	/* Check that the bubbles have actually been loaded. */
	if ((bubbles->p != NULL) && (bubbles->buffer != NULL))
	{
		Mfree(bubbles->buffer);
		Mfree(bubbles->p);
		conf->bubbles = FALSE;
	}
}

/*
 * Display a BubbleGEM help message after a right mouse click.
 */
void display_bubble_mouse(BUBBLES *bubbles, int idx, short xpos, short ypos, USERCONF *conf)
{
	short id	 = 0;
	short msg[8] = {0, 0, 0, 0, 0, 0, 0, 0};

	if (conf->bubbles == TRUE)
	{
		if ((bubbles->buffer != NULL) && (bubbles->p != NULL))
		{
			if ((id = appl_find("BUBBLE  ")) != -1)
			{
				msg[0] = BUBBLEGEM_SHOW;
				msg[1] = _AESapid;
				msg[2] = 0;
				msg[3] = xpos;
				msg[4] = ypos;
				msg[5] = (short)(((int)bubbles->p[idx] >> 16) & 0x0000ffffL);
				msg[6] = (short)(( int)bubbles->p[idx] & 0x0000ffffL);
				msg[7] = 0;
				appl_write(id, 16, msg);
			}
			else {
				form_alert(1, get_free_string(WN_BUBBLES));
			}
		}
	}
}

/*
 * Change the infoline text, according to which button is being moused over.
 * The text string is the same as the BubbleGEM one.
 */
void update_infoline(GEMCLIP *manager, WINDOW *win, OBJECT *tree, BUBBLES *bubbles, short object)
{
	int idx  = 0;
	int skip = FALSE; // Skip writing the infoline at the end of the function if the 'default' case was encountered.

	if (tree == manager->tree)
	{
		switch (object)
		{
		case MANAGER_CURRENT:
			idx = MGRCURR;
			break;
		case MANAGER_VIEW:
			idx = MGRVIEW;
			break;
		case MANAGER_SAVE:
			idx = MGRSAVE;
			break;
		case MANAGER_DELETE:
			idx = MGRDELE;
			break;
		case MANAGER_VIEWMODE_POP:
		case MANAGER_VIEWMODE_EQUIV:
			idx = MGRMODE;
			break;
		case MANAGER_PREVIEW:
			idx = MGRTEXT;
			break;
		case MANAGER_PROPERTIES:
			idx = MGRINFO;
			break;
		case MANAGER_NAME:
			idx = MGRNAME;
			break;
		case MANAGER_DATE:
			idx = MGRDATE;
			break;
		case MANAGER_SIZE:
			idx = MGRSIZE;
			break;
		case MANAGER_ORDER:
			idx = MGRORDR;
			break;
		case MANAGER_CLIP_ICON:
			idx = MGRICON;
			break;
		case MANAGER_CLIPBOARD:
			idx = MGRSCRP;
			break;
		case MANAGER_CLIP01:
		case MANAGER_CLIP02:
		case MANAGER_CLIP03:
		case MANAGER_CLIP04:
		case MANAGER_CLIP05:
		case MANAGER_CLIP06:
		case MANAGER_CLIP07:
		case MANAGER_CLIP08:
		case MANAGER_CLIP09:
		case MANAGER_CLIP10:
		case MANAGER_CLIP11:
		case MANAGER_CLIP12:
			idx = MGRLIST;
			break;
		case MANAGER_PREV1:
		case MANAGER_PREV2:
		case MANAGER_PREV3:
		case MANAGER_PREV4:
		case MANAGER_PREV5:
		case MANAGER_PREV6:
			idx = MGRPREV;
			break;
		default:
			update_clips_infoline(manager); // Rewrite standard infoline.
			skip = TRUE;
			break;
		}
	}
	else if (tree == get_gem_tree(CONFIG)) // Config dialog.
	{
		switch (object)
		{
		case CONFIG_SIDE_MISC:
			idx = CNFMISC;
			break;
		case CONFIG_SIDE_KEYS:
			idx = CNFKEYS;
			break;
		case CONFIG_SIDE_SIZES:
			idx = CNFLIMT;
			break;
		case CONFIG_CYCLING:
			idx = CNFCYCL;
			break;
		case CONFIG_BUBBLES:
			idx = CNFBUBL;
			break;
		case CONFIG_ICONIFY:
			idx = CNFICON;
			break;
		case CONFIG_AUTOICON:
		case CONFIG_AUTOICON_POP:
		case CONFIG_AUTOICON_EQUIV:
			idx = CNFAUTO;
			break;
		case CONFIG_SETTINGS:
			idx = CNFSETT;
			break;
		case CONFIG_WARNINGS:
			idx = CNFWRNG;
			break;
		case CONFIG_CLEANING:
			idx = CNFMPTY;
			break;
		case CONFIG_DB_CLICK:
		case CONFIG_DB_CLICK_EQUIV:
			idx = CNFDBCK;
			break;
		case CONFIG_RMB_CLICK:
		case CONFIG_RMB_CLICK_EQUIV:
			idx = CNFRIGT;
			break;
		case CONFIG_ENTER:
		case CONFIG_ENTER_EQUIV:
			idx = CNFRTRN;
			break;
		case CONFIG_SPACE:
		case CONFIG_SPACE_EQUIV:
			idx = CNFSPAC;
			break;
		case CONFIG_INSERT:
		case CONFIG_INSERT_EQUIV:
			idx = CNFINSR;
			break;
		case CONFIG_SIZE:
			idx = CNFSIZE;
			break;
		case CONFIG_CLIP_SIZE:
			idx = CNFCLIP;
			break;
		case CONFIG_FREE_SPACE:
			idx = CNFFREE;
			break;
		case CONFIG_PATH:
			idx = CNFPATH;
			break;
		case CONFIG_OVERRIDE:
			idx = CNFOVER;
			break;
		case CONFIG_OK:
			idx = CNFOKAY;
			break;
		case CONFIG_CANCEL:
			idx = CNFCNCL;
			break;
		case CONFIG_APPLY:
			idx = CNFAPLY;
			break;
		case CONFIG_DEFAULTS:
			idx = CNFDFLT;
			break;
		case CONFIG_HELP:
			idx = CNFHELP;
			break;
		default:
			/* Rewrite standard infoline according to which tab is displayed. */
			wind_set_str(win->handle, WF_INFO, win->info);
			skip = TRUE;
			break;
		}
	}
	else if (tree == get_gem_tree(INFOS)) // Info dialog.
	{
		switch (object)
		{
		case INFOS_CLOSE:
			idx = ABTOKAY;
			break;
		case INFOS_PREV:
			idx = INFPREV;
			break;
		case INFOS_NEXT:
			idx = INFNEXT;
			break;
		case INFOS_HELP:
			idx = INFHELP;
			break;
		default:
			wind_set_str(win->handle, WF_INFO, win->info); // Rewrite stored infoline.
			skip = TRUE;
			break;
		}
	}

	/* Actually write the new infoline if not already done. */
	if (skip == FALSE) {
		wind_set_str(win->handle, WF_INFO, bubbles->p[idx]);
	}
}
