/*
 * GEMClip, a Clipboard Manager for FreeMiNT and MagiC.
 *
 * Copyright (C) 2011 Jean-Franois Lemaire (jflemaire@skynet.be)
 *
 * This file is part of GEMClip.
 *
 * This program is free software; redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "general.h"

#include "sort.h"

static int compare_clip_order (const void *clip1, const void *clip2);
static int compare_clip_date  (const void *clip1, const void *clip2);
static int compare_clip_size  (const void *clip1, const void *clip2);

/*
 * Reorder clips according to the 3 possible values (order, date, size) and the order
 * (top->bottom or bottom->top). Calls qsort() to do the sorting.
 */
void reorder_listbox(GEMCLIP *manager, int mode)
{
	int i;
	CLIPS clips[manager->nclips];

	if (manager->clipn != NOSCRAP)
	{
		for (i = 0; i < manager->nclips; i++) {
			clips[i] = *manager->clips[i + 1]; // +1 for skipping SCRAP.TXT.
		}

		if (mode == MANAGER_NAME) {
			qsort(clips, manager->nclips, sizeof (CLIPS), compare_clip_order);
		}
		else if (mode == MANAGER_DATE)
		{
			qsort(clips, manager->nclips, sizeof (CLIPS), compare_clip_date);
		}
		else if (mode == MANAGER_SIZE) {
			qsort(clips, manager->nclips, sizeof (CLIPS), compare_clip_size);
		}

		/* Copy back the array of structures. */
		for (i = 0; i < manager->nclips; i++) {
			*manager->clips[i + 1] = clips[i];
		}
	}
}

/*
 * Reverse the order of the clips.
 */
void reverse_order(GEMCLIP *manager)
{
	int i;
	int j;
	CLIPS clips[manager->nclips];

	for (i = 0, j = 1; i < manager->nclips; i++, j++) {
		clips[i] = *manager->clips[j]; // We skip SCRAP.TXT.
	}

	/* Copy back the array of structures. */
	for (i = manager->nclips - 1, j = 1; i >= 0; i--, j++)
	{
		*manager->clips[j] = clips[i];
	}
}

/*
 * Reset the order of the clips. The array of clips is always in the default name/top->bottom order.
 */
void reset_order(GEMCLIP *manager, USERCONF *conf)
{
	int i;
	int j;
	CLIPS clips[manager->nclips];
	_DATETIME time;

	/* If the date is different, the order is currently the default top->down.
	   If it's not, we don't do anything. */
	if ((time.hour   != manager->clips[SCRAP]->time.hour)   ||
	    (time.minute != manager->clips[SCRAP]->time.minute) ||
	    (time.second != manager->clips[SCRAP]->time.second) ||
	    (time.year   != manager->clips[SCRAP]->time.year)   ||
	    (time.month  != manager->clips[SCRAP]->time.month)  ||
	    (time.day	 != manager->clips[SCRAP]->time.day))
	{
		if (conf->order == LIST_BOTTOM)
		{
			for (i = manager->nclips -1, j = 1; i >= 0; i--, j++) {
				clips[i] = *manager->clips[j]; // We skip SCRAP.TXT.
			}
	
			/* Copy back the array of structures. */
			for (i = 0, j = 1; i < manager->nclips; i++, j++)
			{
				*manager->clips[j] = clips[i];
			}
		}
	}
	time = manager->clips[SCRAP]->time;
}

/*
 * Callback comparison function for qsort() that compares 2 strings (name of file).
 */
static int compare_clip_order(const void *clip1, const void *clip2)
{
	const CLIPS *a = (const CLIPS *)clip1;
	const CLIPS *b = (const CLIPS *)clip2;

	return strcmp(a->name, b->name);
}

/*
 * Callback comparison function for qsort() that compares 2 day members of the DATETIME struct.
 */
static int compare_clip_date(const void *clip1, const void *clip2)
{
	const CLIPS *a = (const CLIPS *)clip1;
	const CLIPS *b = (const CLIPS *)clip2;

	if ((a->time.year + 1980) > (b->time.year + 1980)) {
		return 1;
	}
	else if ((a->time.year + 1980) < (b->time.year + 1980)) {
		return -1;
	}
	else
	{
		if (a->time.month > b->time.month) {
			return 1;
		}
		else if (a->time.month < b->time.month) {
			return -1;
		}
		else
		{
			if (a->time.day > b->time.day) {
				return 1;
			}
			else if (a->time.day < b->time.day) {
				return -1;
			}
			else
			{
				if (a->time.hour > b->time.hour) {
					return 1;
				}
				else if (a->time.hour < b->time.hour) {
					return -1;
				}
				else
				{
					if (a->time.minute > b->time.minute) {
						return 1;
					}
					else if (a->time.minute < b->time.minute) {
						return -1;
					}
					else
					{
						if ((a->time.second * 2) > (b->time.second * 2)) {
							return 1;
						}
						else if ((a->time.second * 2) < (b->time.second * 2)) {
							return -1;
						}
						else {
							return 0; // The 2 dates are equal.
						}
					}
				}
			}
		}
	}
}

/*
 * Callback comparison function for qsort() that compares 2 integers (size).
 */
static int compare_clip_size(const void *clip1, const void *clip2)
{
	const CLIPS *a = (const CLIPS *)clip1;
	const CLIPS *b = (const CLIPS *)clip2;

	if (a->size > b->size) {
		return 1;
	}
	else if (a->size < b->size) {
		return -1;
	}
	else {
		return 0;
	}
}

/*
 * Callback comparison function for qsort() that compares 2 integers (clip index).
 */
int compare_clip_index(const void *index1, const void *index2)
{
	const long *a = (const long *)index1;
	const long *b = (const long *)index2;

	if (*a > *b) {
		return 1;
	}
	else if (*a < *b) {
		return -1;
	}
	else {
		return 0;
	}
}
